local M = {}
M.type = "auxiliary"
M.relevantDevice = nil
M.defaultOrder = 1000

local constants = {
    -- Conversion rates
    mpsToMph = 2.23694,
    hpToKw = 0.7457,
    torqueToHpCoef = 7120.54016,
    kWhToJ = 3600000,
    sToH = 1.0 / 3600,
    avToRPM = 9.549296596425384,
    -- Sounds
    inverterBaseVolume = 0.85,
    inverterMaxVolumeKw = 50,
    inverterPitchMinSpeed = 0.25,
    inverterPitchMaxSpeed = 2,
    motorBaseVolume = 0.45,
    motorVolumeMinSpeed = 0.2,
    motorVolumeMaxSpeed = 50,
    motorPitchMinSpeed = 0,
    motorPitchMaxSpeed = 95,
    motorMinPitch = 0.3,
    motorMaxPitch = 3.4
}

local min = math.min
local max = math.max
local abs = math.abs
local mathutil = require("mathutil")
local debugutil = require("debugutil")("hybrid", true)
local maprange = mathutil.maprange
local clamp = mathutil.clamp
local blend = mathutil.blend

local engine = nil
local gearbox = nil
local battery = nil
local esc = nil

local sounds = {
    inverterWhine = nil,
    motorRamp = nil
}

-- Calculated/initial values
local initialWheelBrakeSplit = {}

-- Quick access upvalues
local engineOn = false
local watertemp = 0
local speedMph = 0
local batteryLevel = 1
local totalPowerOutKw = 0
local throttle = 0
local throttleAggressiveness = 0
local gear = "P"

-- Other state variables
local engineDemandReason = "none"
local state = "init"
local nextState = nil
local timeSinceEngineStart = 5
local timeSinceEngineCanShutOff = 0
local timeSinceAppliedBrakes = 0
local startingTime = 0
local isInChargeup = false
local lastThrottle = 0
local chargeThrottle = 0
local disableRegen = false
local timeSinceAbsHit = 0
local timeSinceInit = 0

local function isEngineDemandedOn()
    local maxEvOnlyThrottle

    if gear == "P" then
        maxEvOnlyThrottle = 0.02
    else
        maxEvOnlyThrottle = M.maxEvOnlyThrottle * maprange(speedMph, M.maxEvOnlySpeed / 2, M.maxEvOnlySpeed, 1.0, 0.5)
    end

    if state == "chargemode" then
        engineDemandReason = "charge mode"
        return true
    end
    if not M.enableAutoStop then
        engineDemandReason = "auto stop disabled"
        return true
    end
    if watertemp < 57 then
        engineDemandReason = "temp"
        return true
    end -- engine must be warmed up
    if speedMph > M.maxEvOnlySpeed then
        engineDemandReason = "speed"
        return true
    end -- too fast
    if batteryLevel < M.minBatteryPowerLevel or isInChargeup then
        engineDemandReason = "battery level"
        isInChargeup = true
        return true
    end -- battery too low
    if throttle > maxEvOnlyThrottle or throttleAggressiveness > M.maxEvOnlyThrottleAggressiveness then
        engineDemandReason = "throttle"
        return true
    end
    if engineOn and (input.brake < M.autoStopBrakeThreshold or timeSinceAppliedBrakes < M.autoStopBrakeTime) then
        -- this check only applies when not in P
        if gear ~= "P" then
            engineDemandReason = "brakes"
            return true
        end
    end

    -- Easy checks are done; dynamic checks now

    -- This is the percentage of the max EV-only kW output that's available to us,
    -- based on battery level
    local batteryWorkingRatio = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel)
    local availableEvOnlyPower = batteryWorkingRatio * M.maxEvOnlyPower
    local curPower = gearbox.outputKw

    if engineOn then
        if curPower > 2 then
            engineDemandReason = "accelerating"
            return true
        end -- once engine is on, must coast again to shut off
    else
        if curPower > availableEvOnlyPower or throttle > batteryWorkingRatio * M.evPercentOfTotalPower then
            engineDemandReason = "accelerating"
            return true
        end
    end
    
    -- Put this last so it doesn't override other reasons
    if timeSinceEngineStart < 4 then
        engineDemandReason = "time since start"
        return true
    end -- engine has to run for at least 4 seconds

    engineDemandReason = "none"
    return false -- by default, the engine is off
end

local function updateValues(dt)
    if not gearbox or not engine or not battery then
        return
    end

    engineOn = not (engine.isStalled or engine.ignitionCoef < 0.01 or electrics.values.rpm < engine.idleRPM - 100)
    watertemp = electrics.values.watertemp
    speedMph = electrics.values.wheelspeed * constants.mpsToMph
    batteryLevel = battery.remainingRatio
    throttle = electrics.values.throttleOverride or input.throttle
    throttleAggressiveness = (throttle - lastThrottle) / dt
    gear = electrics.values.gear

    if engineOn then
        timeSinceEngineStart = timeSinceEngineStart + dt
    end

    if M.autoStopBrakeThreshold > 0 and input.brake >= M.autoStopBrakeThreshold and speedMph <= M.maxEvOnlySpeed then
        timeSinceAppliedBrakes = timeSinceAppliedBrakes + dt
    else
        timeSinceAppliedBrakes = 0
    end

    if isInChargeup and batteryLevel >= blend(M.minBatteryPowerLevel, M.maxBatteryPowerLevel, 0.5) then
        isInChargeup = false
    end
end

local function updateState(dt)
    local engineDemandedOn = isEngineDemandedOn()

    debugutil.putValue("Engine demanded on", tostring(engineDemandedOn))
    debugutil.putValue("Engine on reason", engineDemandReason)
    debugutil.putValue("State", state)

    -- keep things like daytime-running-lights on
    electrics.values.running = 1

    if state == "init" then
        timeSinceInit = timeSinceInit + dt

        if timeSinceInit >= 2 then
            if engineOn then
                nextState = "running"
            else
                nextState = "stopped"
            end
        end
    elseif state == "evmode" then
        gearbox:setInputEngaged(false)
        engine.ignitionCoef = 0 -- Kill engine
        engine.isDisabled = true -- Prevent the user from manually restarting the engine
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

        if electrics.values.evMode ~= 1 then
            if M.forceEvMode then
                electrics.values.evMode = 1
            else
                nextState = "init"
            end
        end
    elseif state == "running" then
        local canEngageInGear = gear == "L" or gear == "D"
        local engineInRpmBounds = electrics.values.rpm >= engine.idleRPM
        local canEngineEngage = speedMph >= 2.5 or throttle > 0

        gearbox:setInputEngaged(canEngageInGear and engineInRpmBounds and canEngineEngage)

        if not engineOn then
            -- engine died?
            
            timeSinceInit = 0
            nextState = "init"
        elseif electrics.values.evMode == 1 then
            nextState = "evmode"
        elseif electrics.values.chargeMode == 1 and gear == "P" then
            chargeThrottle = 0
            nextState = "chargemode"
        else
            if engineDemandedOn then
                timeSinceEngineCanShutOff = 0
            else
                if gear ~= "N" then -- engine keeps state in N
                    timeSinceEngineCanShutOff = timeSinceEngineCanShutOff + dt
                    
                    if timeSinceEngineCanShutOff > 1 then
                        gearbox:setInputEngaged(false)
                        engine.ignitionCoef = 0 -- Kill engine
                        engine.isDisabled = true -- Prevent the user from manually restarting the engine
                        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)

                        nextState = "stopped"
                    end
                end
            end
        end
    elseif state == "chargemode" then
        local inverseRelativePowerLevel = maprange(batteryLevel, 0.9, 0.95, 1.0, 0.0)
        local maxFastChargePower = M.maxFastChargePower or mathutil.motorTorqueToKw(engine.outputTorque1, electrics.values.rpm)
        local curGenOutput = -min(totalPowerOutKw, 0)
        local targGenOutput = inverseRelativePowerLevel * M.maxFastChargePower
        local error = (targGenOutput - curGenOutput) / targGenOutput

        chargeThrottle = clamp(chargeThrottle + (error * dt * M.chargeThrottleIFactor), 0, 1)

        if gear ~= "P" or electrics.values.evMode == 1 or batteryLevel >= 0.93 then
            electrics.values.chargeMode = 0
        end

        if electrics.values.chargeMode ~= 1 then
            nextState = "init"
        end
    elseif state == "stopped" then
        timeSinceEngineCanShutOff = 0
        electrics.values.checkengine = false -- Keep this from coming on (usually tied to engine.isDisabled)
        electrics.values.iceThrottle = 0
        gearbox:setInputEngaged(false)

        if electrics.values.evMode == 1 then
            nextState = "evmode"
        elseif gear ~= "N" and (engineDemandedOn or electrics.values.chargeMode == 1) then
            engine.isDisabled = false -- allow engine to restart

            if electrics.values.rpm < engine.idleRPM then
                engine:activateStarter() -- start engine
                nextState = "startup"
                timeSinceEngineStart = 0
                startingTime = 0
            else
                engine.ignitionCoef = 1 -- engine already spinning, engage ignition
                nextState = "running"
            end
        end
    elseif state == "startup" then
        electrics.values.iceThrottle = 0
        startingTime = startingTime + dt

        if engineOn then
            if timeSinceEngineStart > 0.15 then
                engine:deactivateStarter()
                nextState = "running"
            end
        else
            if startingTime > 3 then
                -- Something's not right; retry at init

                engine:deactivateStarter()
                timeSinceInit = 0
                nextState = "init"
            end
        end
    end
end

local function updateBattery(dt)
    totalPowerOutKw = M.idlePowerDraw

    local motorKw = gearbox.outputKw

    if motorKw > 0 then
        totalPowerOutKw = totalPowerOutKw + motorKw
    else
        totalPowerOutKw = totalPowerOutKw + (motorKw * M.regenEfficiency)
    end

    if engineOn then
        local requestedGenTorque = 0
        local relativePowerLevel = 0

        if state == "chargemode" then
            relativePowerLevel = maprange(batteryLevel, 0.9, 1.0)
        else
            relativePowerLevel = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel)
        end
        
        local inverseRelativePowerLevel = 1 - relativePowerLevel

        local maxGenCoef = 0

        if state == "chargemode" then
            maxGenCoef = maprange(electrics.values.rpm, engine.idleRPM, M.engineMaxGeneratorRpm, 0, 1)
        elseif gear == "P" then
            maxGenCoef = 1
        else
            maxGenCoef = inverseRelativePowerLevel
        end
        
        if battery.jbeamData.maxPowerInput then
            local batteryPowerInputCoef = maprange(batteryLevel, 0.9, 0.95, 1, 0)
            local batteryPowerInputAvailable = battery.jbeamData.maxPowerInput * batteryPowerInputCoef
            local maxEngineKw = mathutil.motorTorqueToKw(M.maxEngineGeneratorTorque, M.engineMaxGeneratorRpm)
            
            -- Restrict generator torque demand based on how much power the battery
            -- can currently accept
            maxGenCoef = min(maxGenCoef, batteryPowerInputAvailable / maxEngineKw)
        end

        local rpmCoef = maprange(electrics.values.rpm, 0, M.engineMaxGeneratorRpm * 0.9)
        local genCoef = maprange(rpmCoef, 0, inverseRelativePowerLevel, inverseRelativePowerLevel, maxGenCoef)

        requestedGenTorque = M.maxEngineGeneratorTorque * genCoef
        gearbox:setEngineGenerationTorque(requestedGenTorque)
        debugutil.putValue("Inverse Relative Power Level", inverseRelativePowerLevel)
        debugutil.putValue("Requested Generator Torque", requestedGenTorque)
        totalPowerOutKw = totalPowerOutKw - gearbox.generatorKw

        --electrics.values.lockupClutchRatio = maprange(electrics.values.rpm, engine.idleRPM, M.engineMaxGeneratorRpm, 0, genCoef)
    end

    local totalPowerOutKj = totalPowerOutKw * dt

    debugutil.putValue("Battery power (kW)", totalPowerOutKw)
    debugutil.putValue("Generator power (kW)", gearbox.generatorKw)

    battery.storedEnergy = mathutil.clamp(battery.storedEnergy - totalPowerOutKj * 1000, 0, battery.initialStoredEnergy)
end

local function updateSounds()
    -- Update sounds
    local inverterPitchCoef = maprange(speedMph, constants.inverterPitchMinSpeed, constants.inverterPitchMaxSpeed)
    local inverterEffort = math.abs(gearbox.outputKw) / constants.inverterMaxVolumeKw
    local inverterVolume = constants.inverterBaseVolume * mathutil.clamp(inverterEffort, 0.0, 1.0)
    local inverterPitch = mathutil.clamp(inverterPitchCoef, 0.4, 1.0)

    obj:setVolumePitch(sounds.inverterWhine, inverterVolume, inverterPitch)

    local motorPitch =
        maprange(
        speedMph,
        constants.motorPitchMinSpeed,
        constants.motorPitchMaxSpeed,
        constants.motorMinPitch,
        constants.motorMaxPitch,
        false
    )
    local motorVolume =
        maprange(
        speedMph,
        constants.motorVolumeMinSpeed,
        constants.motorVolumeMaxSpeed,
        0,
        constants.motorBaseVolume,
        false
    )

    obj:setVolumePitch(sounds.motorRamp, motorVolume, motorPitch)
end

local function updateGUI(dt)
    local str = {
        soc = batteryLevel,
        kw = totalPowerOutKw,
        maxKw = gearbox and gearbox.maxMotorPower or 0
    }

    gui.send("hybrid", str)
    debugutil.sendStreams()
end

local function updateGFX(dt)
    updateValues(dt)

    local evThrottle = 0.0
    local iceThrottle = 0.0

    if gearbox and engine and battery then
        updateState(dt)
        updateBattery(dt)
        
        gearbox.maxPowerOverride = nil

        local actualMotorPercentMaxTorque = 0.0
        local evThrottleFactor = nil
        local gearboxMaxMotorTorque = gearbox:getMaxMotorTorque()
        local gearboxAdjustedMaxMotorTorque = gearboxMaxMotorTorque * gearbox.motorGearRatio
        local actualSpeedSign = mathutil.sign(electrics.values.avgWheelAV)
        local desiredSpeedSign = gear == "R" and -1 or 1
        local batteryWorkingRatio = 0

        if state == "evmode" then
            batteryWorkingRatio = maprange(batteryLevel, 0.15, M.minBatteryPowerLevel)
        else
            batteryWorkingRatio = maprange(batteryLevel, M.minBatteryPowerLevel, M.maxBatteryPowerLevel)
        end
        
        local availableBatteryPower = batteryWorkingRatio * ((battery.jbeamData.maxPowerOutput or gearbox.maxMotorPower) - M.idlePowerDraw)
        local genPowerAvailableToMotor = maprange(batteryWorkingRatio, 0, 1, 0.5, 1) * max(gearbox.generatorKw, 0)
        local availableEvPower = availableBatteryPower + genPowerAvailableToMotor
        local availableEvThrottle = 1
        local gearboxMotorSpeedRpm = abs(gearbox.motorAV * constants.avToRPM)
        local gearboxMaxAdjustedMotorKw = mathutil.motorTorqueToKw(gearboxMaxMotorTorque, gearboxMotorSpeedRpm)
        local maxRegenPctOfMaxMotorPower = 1
        
        if battery.jbeamData.maxPowerInput then
            local batteryPowerInputCoef = maprange(batteryLevel, 0.9, 0.95, 1, 0)
            local batteryPowerInputAvailable = battery.jbeamData.maxPowerInput * batteryPowerInputCoef
            
            -- Calculate what percentage of regen we can use, taking into account
            -- the maximum amount of power the battery can currently accept, minus
            -- however much power the engine is putting into the battery
            maxRegenPctOfMaxMotorPower = (batteryPowerInputAvailable - gearbox.generatorKw) / (gearboxMaxAdjustedMotorKw * M.regenEfficiency)
        end
        
        local maxEffectiveRegenCoef = min(M.maxRegenCoef, maxRegenPctOfMaxMotorPower)
        local maxRegen = maxEffectiveRegenCoef * maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed)
        local throttleOffset = max(maxRegen * 0.25, 0.01)
    
        -- Fill in some debug values
        debugutil.putValue("Battery SOC % (usable)", math.floor(batteryWorkingRatio * 100))
        debugutil.putValue("Available EV power (kW)", availableEvPower)
        debugutil.putValue("Throttle % EV thresh.", math.floor(throttle * 100 / M.maxEvOnlyThrottle))

        if input.brake > 0 then
            -- Launch control
            if throttle > 0 then
                iceThrottle = maprange(throttle, 0, 0.2, 0, 0.4, true)
            end

            input.throttle = 0
            throttle = 0
            lastThrottle = 0
            throttleAggressiveness = 0

            if gear == "N" or gear == "P" then
                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Reset brake split coef
                        wheel.brakeInputSplit = initialWheelBrakeSplit[i][1]
                        wheel.brakeSplitCoef = initialWheelBrakeSplit[i][2]
                    end
                end
                
                debugutil.putValue("Friction % of full brakes", math.floor(input.brake * 100))
            else
                local maxBrakeTorque = 0.0
                local absActive = false

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wheel = wheels.wheels[i]

                    if wheel then
                        -- Find total braking torque
                        if wheel.brakeTorque > maxBrakeTorque then
                            maxBrakeTorque = wheel.brakeTorque
                        end

                        -- Determine if ABS active
                        if wheel.absActive then
                            absActive = true
                        end
                    end
                end

                if disableRegen then
                    if not absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 1.0 then
                            disableRegen = false
                            timeSinceAbsHit = 0
                        end
                    end
                else
                    if absActive then
                        timeSinceAbsHit = timeSinceAbsHit + dt

                        if timeSinceAbsHit > 0.5 then
                            disableRegen = true
                            timeSinceAbsHit = 0
                        end
                    else
                        timeSinceAbsHit = 0
                    end
                end

                local regenBrakePedal = maprange(input.brake, 0, M.regenPercentOfBrakingForce)
                local regenAccelPedal = gear == "L" and maprange(throttle, throttleOffset, 0.0) or 0
                local posAccelPedal = maprange(throttle, throttleOffset, 1)
                local regen = clamp(regenBrakePedal + regenAccelPedal, 0, 1)

                evThrottle = -(regen * maxRegen) + min(posAccelPedal, 0.15) -- launch control
                evThrottleFactor = 100 -- faster for brake application

                local actualRegenCoef = evThrottle > -0.01 and 0 or clamp((gearbox.motorOutputTorque * actualSpeedSign) / (evThrottle * gearboxAdjustedMaxMotorTorque), 0, 1)
                local actualEffectiveRegen = actualRegenCoef * maxRegen * M.regenPercentOfBrakingForce
                local brakesCoveredByRegen = maprange(input.brake, actualEffectiveRegen, 1, 1, 0)
                local neededFrictionBrake = maprange(input.brake + (regenAccelPedal * M.regenPercentOfBrakingForce), brakesCoveredByRegen, 1)

                debugutil.putValue("Max Regen", maxRegen)
                debugutil.putValue("Regen % of full brakes", math.floor(actualEffectiveRegen * 100))
                debugutil.putValue("Regen % of full regen", math.floor(regen * 100))
                debugutil.putValue("Friction % of full brakes", math.floor(neededFrictionBrake * 100))

                for i = 0, wheels.wheelCount - 1, 1 do
                    local wd = wheels.wheels[i]

                    wd.brakeInputSplit = 1 - brakesCoveredByRegen
                    
                    if wd.isPropulsed then
                        -- Propulsed wheels are affected by regenerative braking
                        
                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * max(neededFrictionBrake - actualRegenCoef, 0)
                    else
                        -- Unpropulsed wheels are not
                        
                        wd.brakeSplitCoef = initialWheelBrakeSplit[i][2] * neededFrictionBrake
                    end
                end
            end
        else
            disableRegen = false

            local throttleFactor = electrics.values.throttleFactor or 1.0
            
            actualMotorPercentMaxTorque = 1.0 - maprange(throttle, availableEvThrottle, 1.01, 0.0, 1.0, true)

            local iceThrottleSource = 0

            if gear == "L" then
                -- Regenerative braking when letting off accelerator
                local posThrottle = maprange(throttle, throttleOffset, 1.0, 0.0, 1.0, true)
                local negThrottle = maprange(throttle, throttleOffset, 0.0, 0.0, 1.0, true)
                local posEvThrottle = maprange(posThrottle, 0, M.evPercentOfTotalPower * max(availableEvThrottle, 0.01)) * availableEvThrottle
                local negEvThrottle = negThrottle * maxRegen

                evThrottle = posEvThrottle - negEvThrottle
                iceThrottleSource = posThrottle
            else
                local minEvThrottle = maprange(speedMph, M.minRegenFadeSpeed, M.maxRegenFadeSpeed, 0.0, M.evCoastThrottle)

                evThrottle = maprange( throttle, 0, M.evPercentOfTotalPower * max(availableEvThrottle, 0.01), minEvThrottle, availableEvThrottle )
                iceThrottleSource = throttle
            end

            local iceThrottleThreshold = batteryWorkingRatio * M.evPercentOfTotalPower

            evThrottleFactor = maprange( throttleAggressiveness, M.minThrottleAggressivenessRamp, M.maxThrottleAggressivenessRamp, M.minThrottleAggressivenessFactor, M.maxThrottleAggressivenessFactor )
            iceThrottle = maprange(iceThrottleSource, iceThrottleThreshold, 1, 0, 1)
        end
        
        local positiveSpeedCreep = maprange(speedMph, 0, 5, M.evCreepThrottle, 0)
        local hillAssistCreep = maprange(speedMph * actualSpeedSign * desiredSpeedSign, 0, -1.5, 0, M.maxHillAssistThrottle)
        local creep = max(positiveSpeedCreep, hillAssistCreep)
            
        if actualSpeedSign ~= desiredSpeedSign and throttle < 0.01 then
            evThrottleFactor = 100
        end

        if input.brake < 0.25 and gear ~= "P" and gear ~= "N" then
            evThrottle = evThrottle + creep
        end

        gearbox:setElectricTorqueDemand(evThrottle, evThrottleFactor)
        debugutil.putValue("Motor % of max torque", actualMotorPercentMaxTorque)
        
        if evThrottle > 0 then
            gearbox.maxPowerOverride = min(gearbox.maxMotorPower, availableEvPower)
        else
            gearbox.maxPowerOverride = nil
        end

        local motorPowerPercent = gearbox.outputKw / gearbox.maxMotorPower

        if gear == "N" then
            iceThrottle = throttle
--        elseif isInChargeup then
--            iceThrottle = max(iceThrottle, M.engineGeneratorMaxMinThrottle)
        else
            if gear == "D" and input.brake > 0.05 and speedMph < 0.5 then
                iceThrottle = min(throttle, 0.05)
            else
                local maxMinThrottle =
                    maprange(
                    batteryLevel,
                    M.minBatteryPowerLevel,
                    M.maxBatteryPowerLevel,
                    M.engineGeneratorMaxMinThrottle,
                    0
                )
            
                iceThrottle = max(iceThrottle, maxMinThrottle)
            end
        end

        if state == "running" then
            if gear == "R" then
                electrics.values.iceThrottle = min(iceThrottle, 0.05)
            else
                electrics.values.iceThrottle = iceThrottle
            end
        elseif state == "chargemode" then
            electrics.values.iceThrottle = chargeThrottle
        end

        if not gearbox.inputEngaged then
            -- In order to prevent over-reving when the ICE engages,
            -- we have to predict a close gear ratio range and keep
            -- the transmission in that range while the ICE is disengaged
            local estGearRatio =
                maprange(speedMph, M.minGearRatioSpeed, M.maxGearRatioSpeed, gearbox.maxGearRatio, gearbox.minGearRatio)

            gearbox:setGearRatio(estGearRatio)
        end
    end

    if battery then
        debugutil.putValue("Battery SOC %", tostring(battery.remainingRatio * 100))
    end

    debugutil.putValue("Input Throttle", (throttle or 0))
    debugutil.putValue("ICE Throttle", iceThrottle)
    debugutil.putValue("EV Throttle", evThrottle)
    debugutil.putValue("Throttle Aggr.", throttleAggressiveness)

    if nextState then
        state = nextState
    end

    lastThrottle = throttle

    debugutil.putValue("ICE Torque", engine.outputTorque1)
    debugutil.putValue("Charge Throttle", chargeThrottle)
    debugutil.putValue("Motor kW", gearbox.outputKw)
    debugutil.putValue("Motor RPM", gearbox.outputAV1 * 9.549)
    debugutil.putValue("Motor max out torque", gearbox:getMaxMotorTorque() * gearbox.motorGearRatio)
    debugutil.putValue("Motor out torque", gearbox.motorOutputTorque)
    debugutil.putValue("Gearbox gear ratio", gearbox.gearRatio)

    updateGUI(dt)
    updateSounds()
end

local function init(jbeamData)
    M.jbeamData = jbeamData
    M.enableAutoStop = jbeamData.enableAutoStop ~= false
    M.maxEvOnlyPower = jbeamData.maxEvOnlyPower or 40 -- kW
    M.maxEvOnlySpeed = jbeamData.maxEvOnlySpeed or 45 -- MPH
    M.maxEvOnlyThrottle = jbeamData.maxEvOnlyThrottle or 0.45 -- ratio
    M.maxEvOnlyThrottleAggressiveness = jbeamData.maxEvOnlyThrottleAggressiveness or 20
    M.minGearRatioSpeed = jbeamData.minGearRatioSpeed or 10
    M.maxGearRatioSpeed = jbeamData.maxGearRatioSpeed or 90
    M.minBatteryPowerLevel = 0.25 -- ratio; low end of max-kW slope
    M.maxBatteryPowerLevel = 0.7 -- ratio; high end of max-kW slope
    M.minRegenFadeSpeed = jbeamData.minRegenFadeSpeed or 4
    M.maxRegenFadeSpeed = jbeamData.maxRegenFadeSpeed or 14
    M.maxEngineGeneratorTorque = jbeamData.maxEngineGeneratorTorque or 180
    M.maxFastChargePower = jbeamData.maxFastChargePower
    M.autoStopBrakeThreshold = jbeamData.autoStopBrakeThreshold or 0
    M.autoStopBrakeTime = jbeamData.autoStopBrakeTime or 0
    M.engineMaxGeneratorRpm = jbeamData.maxGeneratorRpm or 0
    M.engineGeneratorMaxMinThrottle = jbeamData.engineGeneratorMaxMinThrottle or 0.15
    M.minThrottleAggressivenessRamp = jbeamData.minThrottleAggressivenessRamp or 2
    M.maxThrottleAggressivenessRamp = jbeamData.maxThrottleAggressivenessRamp or 15
    M.minThrottleAggressivenessFactor = jbeamData.minThrottleAggressivenessFactor or 1
    M.maxThrottleAggressivenessFactor = jbeamData.maxThrottleAggressivenessFactor or 10
    M.maxRegenCoef = jbeamData.maxRegenCoef or 0.6
    M.regenEfficiency = jbeamData.regenEfficiency or 0.93
    M.regenPercentOfBrakingForce = jbeamData.regenPercentOfBrakingForce or 0.3
    M.evPercentOfTotalPower = jbeamData.evPercentOfTotalPower or 0.65
    M.idlePowerDraw = jbeamData.idlePowerDraw or 1 -- kW
    M.evCreepThrottle = jbeamData.evCreepThrottle or 0.18
    M.evCoastThrottle = jbeamData.evCoastThrottle or -0.05
    M.maxHillAssistThrottle = jbeamData.maxHillAssistThrottle or 1.0
    M.chargeThrottleIFactor = jbeamData.chargeThrottleIFactor or 2.75
    M.startEvMode = jbeamData.startEvMode == true
    M.forceEvMode = jbeamData.forceEvMode == true

    torqueDemandSmoother =
        newTemporalSmoothingNonLinear(constants.torqueDemandSmoothingIn, constants.torqueDemandSmoothingOut)

    -- Pre-fill debug table with empty strings to set up proper order
    debugutil.clearValues()
    debugutil.putValue("State", "")
    debugutil.putValue("Engine demanded on", "")
    debugutil.putValue("Engine on reason", "")
    debugutil.putValue("Battery power (kW)", "")
    debugutil.putValue("Motor kW", "")
    debugutil.putValue("Generator power (kW)", "")
    debugutil.putValue("Inverse Relative Power Level", "")
    debugutil.putValue("Requested Generator Torque", "")
    debugutil.putValue("Battery SOC %", "")
    debugutil.putValue("Battery SOC % (usable)", "")
    debugutil.putValue("Available EV power (kW)", "")
    debugutil.putValue("Throttle % EV thresh.", "")
    debugutil.putValue("Throttle Aggr.", "")
    debugutil.putValue("Input Throttle", "")
    debugutil.putValue("Charge Throttle", "")
    debugutil.putValue("ICE Throttle", "")
    debugutil.putValue("EV Throttle", "")
    debugutil.putValue("Max Regen", "")
    debugutil.putValue("Regen % of full brakes", "")
    debugutil.putValue("Regen % of full regen", "")
    debugutil.putValue("Friction % of full brakes", "")
    debugutil.putValue("Motor max out torque", "")
    debugutil.putValue("Motor out torque", "")
    debugutil.putValue("Motor % of max torque", "")
    debugutil.putValue("ICE Torque", "")
    debugutil.putValue("Motor RPM", "")
    debugutil.putValue("Gearbox gear ratio", "")
end

local function reset()
    timeSinceInit = 0
    state = "init"
    nextState = nil

    init(M.jbeamData)
    M.initSecondStage()
end

local function initSounds()
    sounds.inverterWhine =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/inverter.ogg",
        "AudioDefaultLoop3D",
        "inverterWhine",
        engine.engineNodeID
    )

    sounds.motorRamp =
        obj:createSFXSource(
        "vehicles/hybrid_common/sounds/motor_ramp.ogg",
        "AudioDefaultLoop3D",
        "motorRamp",
        engine.engineNodeID
    )
end

local function initSecondStage()
    engine = powertrain.getDevice("mainEngine")
    gearbox = powertrain.getDevice("gearbox")
    battery = energyStorage.getStorage("tractionBattery")

    if not gearbox or gearbox.type ~= "ecvtGearbox" then
        gearbox = nil
    end

    if battery then
        M.minBatteryPowerLevel = battery.jbeamData.minWorkingSoc or M.minBatteryPowerLevel
        M.maxBatteryPowerLevel = battery.jbeamData.maxWorkingSoc or M.maxBatteryPowerLevel
        
        if not M.maxFastChargePower then
            M.maxFastChargePower = battery.jbeamData.maxPowerInput
        end
    end

    if engine then
        if not M.jbeamData.maxGeneratorRpm then
            M.engineMaxGeneratorRpm = engine:getTorqueData().maxPowerRPM
        end

        if not M.jbeamData.maxFastChargePower then
            -- Calculate a better default for this

            M.maxFastChargePower =
                M.engineMaxGeneratorRpm * constants.hpToKw * (M.maxEngineGeneratorTorque / constants.torqueToHpCoef)
        end
    end

    for i = 0, wheels.wheelCount - 1, 1 do
        if not initialWheelBrakeSplit[i] then
            initialWheelBrakeSplit[i] = {wheels.wheels[i].brakeInputSplit, wheels.wheels[i].brakeSplitCoef}
        end
    end

    if M.startEvMode then
        electrics.values.evMode = 1
    end
end

M.updateGFX = updateGFX
M.init = init
M.initSounds = initSounds
M.initSecondStage = initSecondStage
M.reset = reset

return M
